from jija.forms.validators import *
from jija.forms.exceptions import ValidationError


class Field:
    validators = ()

    def __init__(self, *, required=True, default=None):
        self.required = required
        self.default = default

    async def validate(self, value):
        if not value:
            if self.required and self.default is None:
                print(self.default)
                raise ValidationError('Обязательное поле', value)
            else:
                return value or self.default

        for validator in self.validators:
            value = await validator.validate(value, self)

        return value


class CharField(Field):
    validators = (LengthMinValidator, LengthMaxValidator)

    def __init__(self, *, min_length=None, max_length=None, regex=None, **kwargs):
        super().__init__(**kwargs)
        self.min_length = min_length
        self.max_length = max_length
        self.regex = regex


class NumericField(Field):
    def __init__(self, *, min_value=None, max_value=None, **kwargs):
        super().__init__(**kwargs)
        self.min_value = min_value
        self.max_value = max_value


class IntegerField(NumericField):
    validators = (IntegerValidator, RangeMinValidator, RangeMaxValidator)


class FloatField(NumericField):
    validators = (FloatValidator, RangeMinValidator, RangeMaxValidator)


class DateField(Field):
    validators = (DateValidator, RangeMinValidator, RangeMaxValidator)

    def __init__(self, *, min_value=None, max_value=None, **kwargs):
        super().__init__(**kwargs)
        self.min_value = min_value
        self.max_value = max_value


class SelectField(Field):
    validators = (OptionsValidator,)

    def __init__(self, *, options, **kwargs):
        super().__init__(**kwargs)
        self.options = options



import inspect

from jija.forms import fields
from jija.forms.exceptions import ValidationError


class Form:
    def __init__(self, data=None):
        self.data = data or {}
        self.errors = {}

        self.__valid = None

    async def validate(self):
        for name, obj in self.__class__.__dict__.items():
            if issubclass(type(obj), fields.Field):
                try:
                    self.data[name] = await obj.validate(self.data.get(name))

                except ValidationError as exception:
                    self.errors[name] = exception.error
                    self.data[name] = exception.value

        if inspect.iscoroutinefunction(self.clean):
            await self.clean()
        else:
            self.clean()

        self.__valid = len(self.errors) == 0
        if not self.__valid:
            print(self.errors)

        return self.__valid

    @property
    def valid(self):
        return self.__valid

    def clean(self):
        pass



from jija.config import base, fields

from pathlib import Path as Path
import sys


class StructureConfig(base.Config):
    PROJECT_PATH = None
    CORE_PATH = None
    APPS_PATH = None
    PYTHON_PATH = fields.CharField(default=sys.executable)

    def __init__(self, *, project_path=None, core_dir='core', apps_dir='apps', python_path=None):
        super().__init__(project_path=project_path, core_dir=core_dir, apps_dir=apps_dir, python_path=python_path)

    @classmethod
    async def validate(cls, values):
        project_path = cls.__get_project_path(values['project_path'])
        return {
            'PROJECT_PATH': project_path,
            'CORE_PATH': project_path.joinpath(values['core_dir']),
            'APPS_PATH': project_path.joinpath(values['apps_dir']),
            'PYTHON_PATH': await StructureConfig.PYTHON_PATH.validate(values['python_path'])
        }

    @staticmethod
    def __get_project_path(project_path):
        if isinstance(project_path, Path):
            if not project_path.is_absolute():
                return project_path.absolute()

            return project_path

        path = Path(project_path) if project_path else Path()
        return path.absolute()



from aiohttp import web
import asyncio

from pathlib import Path
from typing import List
import importlib
import sys
import os

from jija.utils.collector import collect_subclasses
from jija.config import base as config_base
from jija import middlewares
from jija import commands
from jija.app import App
from jija import config


class AppGetter(type):
    def __getattr__(self, item):
        """
        :type item: str
        :rtype: App
        """

        app = Apps.apps.get(item)
        if app:
            return app

        raise AttributeError(item)


class Apps(metaclass=AppGetter):
    apps = {}
    commands = {
        'system': commands.COMMANDS
    }

    __REQUIRED_CONFIGS = {
        config.StructureConfig,
        config.DriversConfig,
        config.NetworkConfig
    }

    __INITED_CONFIGS: List['config_base.Config'] = []
    __PREFLIGHT_TASKS = []

    @classmethod
    def load(cls):
        cls.__init_configs()
        Apps.apps['core'] = cls.__create_base_app()
        cls.__collect(config.StructureConfig.APPS_PATH, Apps.apps['core'])
        cls.__register_apps()

    @classmethod
    def config_init_callback(cls, config_class):
        cls.__INITED_CONFIGS.append(config_class)

    @classmethod
    def __init_configs(cls):
        for config_class in cls.__REQUIRED_CONFIGS:
            if config_class not in cls.__INITED_CONFIGS:
                config_class()

        asyncio.get_event_loop().run_until_complete(cls.__freeze_configs())

    @classmethod
    async def __freeze_configs(cls):
        for config_class in cls.__INITED_CONFIGS:
            await config_class.freeze()
            cls.__PREFLIGHT_TASKS.append(config_class.preflight)

    @classmethod
    def __create_base_app(cls):
        """
        :rtype: web.Application
        """

        aiohttp_app = web.Application()

        aiohttp_app.middlewares.extend([
            middlewares.print_request.PrintRequest(),
        ])

        if cls.app_exists(config.StructureConfig.CORE_PATH):
            app_class = cls.get_modify_class(config.StructureConfig.CORE_PATH)
        else:
            app_class = App

        for config_unit in cls.__INITED_CONFIGS:
            aiohttp_app = config_unit.base_app_update(aiohttp_app)

        app = app_class(path=config.StructureConfig.CORE_PATH, aiohttp_app=aiohttp_app, name='core')

        return app

    @staticmethod
    def app_exists(path: Path) -> bool:
        return path.joinpath('app.py').exists()

    @classmethod
    def __collect(cls, path: Path, parent: App):
        if not path.exists():
            return

        for sub_app_name in os.listdir(path):
            sub_app_name: str

            next_path = path.joinpath(sub_app_name)
            if App.is_app(next_path):
                app = cls.get_modify_class(next_path)(path=next_path, parent=parent, name=sub_app_name)
                cls.commands[sub_app_name] = app.commands
                cls.apps[sub_app_name] = app
                cls.__collect(path.joinpath(sub_app_name), app)

    @staticmethod
    def get_modify_class(path: Path) -> type:
        modify_class_path = path.joinpath('app')
        import_path = ".".join(modify_class_path.relative_to(config.StructureConfig.PROJECT_PATH).parts)

        module = importlib.import_module(import_path)
        modify_class = list(collect_subclasses(module, App))
        return modify_class[0] if modify_class else App

    @classmethod
    def __register_apps(cls, app=None):
        """
        :type app: App
        """

        if not app:
            app = cls.apps['core']

        if not app.childes and app.name != 'core':
            app.parent.aiohttp_app.add_subapp(f'/{app.name}', app.aiohttp_app)

        for child in app.childes:
            cls.__register_apps(child)

    @classmethod
    def get_command(cls, module, command):
        """
        :type module: str
        :type command: str
        :rtype: type
        """

        if module is None:
            module = 'system'

        return cls.commands[module][command]

    @classmethod
    def run_command(cls):
        args = sys.argv
        command = args[1].split('.')
        Apps.load()
        asyncio.get_event_loop().run_until_complete(cls.__preflight())

        if len(command) == 1:
            module = None
            command = command[0]
        else:
            module, command = command

        command_class = cls.get_command(module, command)
        command_obj = command_class()
        command_obj.run()

    @classmethod
    async def __preflight(cls):
        for task in cls.__PREFLIGHT_TASKS:
            await task()



from jija.config import base, fields

from aiohttp_session.cookie_storage import EncryptedCookieStorage
import aiohttp_session


class AuthConfig(base.Config):
    SECRET_KEY: str = fields.CharField()

    def __init__(self, *, secret_key):
        super().__init__(secret_key=secret_key)

    @classmethod
    def base_app_update(cls, aiohttp_app):
        aiohttp_session.setup(aiohttp_app, EncryptedCookieStorage(cls.SECRET_KEY))
        return aiohttp_app



import importlib
import os
from pathlib import Path

from aiohttp import web
from jija import config

from jija.middleware import Middleware
from jija.utils.collector import collect_subclasses
from jija.command import Command
from jija import router


class App:
    def __init__(self, *, name, path, aiohttp_app=None, parent=None):
        """
        :type name: str
        :type path: jija.utils.path.Path
        :type aiohttp_app: web.Application
        :type parent: App
        """

        self.__parent = parent
        if parent:
            parent.add_child(self)

        self.__path = path
        self.__name = name
        self.__is_core = aiohttp_app is not None

        self.__router = None
        self.__middlewares = None
        self.__commands = None

        self.__aiohttp_app = None

        self.__childes = []
        self.__load(aiohttp_app)

    @property
    def parent(self):
        """:rtype: App"""
        return self.__parent

    @property
    def name(self):
        """:rtype: str"""
        return self.__name

    @property
    def router(self) -> router.Router:
        return self.__router

    @property
    def middlewares(self):
        """:rtype: list[Middleware]"""
        return self.__middlewares

    @property
    def aiohttp_app(self):
        """:rtype: web.Application"""
        return self.__aiohttp_app

    @property
    def childes(self):
        """:rtype: list[App]"""
        return self.__childes

    @property
    def path(self) -> Path:
        return self.__path

    @property
    def commands(self) -> list:
        return self.__commands

    def __load(self, aiohttp_app: web.Application = None):
        self.__router = self.__get_router()
        self.__middlewares = self.__get_middlewares()
        self.__commands = self.__get_commands()

        self.__aiohttp_app = self.get_aiohttp_app(aiohttp_app)

    def __get_router(self) -> "router.Router":
        if not self.exist('routes'):
            raw_routes = []

        else:
            import_path = self.get_import_path('routes')
            routes_module = importlib.import_module(import_path)

            raw_routes = getattr(routes_module, 'routes', [])

        app_router = router.Router(raw_routes)
        return app_router

    def __get_middlewares(self) -> list:
        if not self.exist('middlewares'):
            return []

        import_path = self.get_import_path('middlewares')
        middlewares_module = importlib.import_module(import_path)
        middlewares = collect_subclasses(middlewares_module, Middleware)

        return list(map(lambda middleware: middleware(), middlewares))

    def __get_commands(self) -> dict:
        if not self.exist('commands'):
            return {}

        commands = {}
        commands_path = self.__path.joinpath('commands')
        for file in os.listdir(commands_path):
            if file.endswith('.py') and not file.startswith('_'):

                import_path = self.get_import_path(f'commands.{file.removesuffix(".py")}')
                command_module = importlib.import_module(import_path)

                command = list(collect_subclasses(command_module, Command))
                if command:
                    commands[file.replace('.py', '')] = command[0]

        return commands

    @staticmethod
    def is_app(path: Path) -> bool:
        if not path.is_dir() or not path.joinpath('app.py').exists():
            return False

        for part in path.parts:
            if part.startswith('__'):
                return False

        return True

    def get_aiohttp_app(self, aiohttp_app: web.Application = None) -> web.Application:
        aiohttp_app = aiohttp_app or web.Application()

        aiohttp_app.middlewares.extend(self.__middlewares)
        aiohttp_app.add_routes(self.__router.routes)

        return aiohttp_app

    def add_child(self, child: "App"):
        self.__childes.append(child)

    def get_import_path(self, to: str) -> str:
        modify_class_path = self.path.joinpath(to)
        return ".".join(modify_class_path.relative_to(config.StructureConfig.PROJECT_PATH).parts)

    def exist(self, name: str) -> bool:
        return os.path.exists(f'/{self.__path}/{name}') or os.path.exists(f'/{self.__path}/{name}.py')



from jija.command import Command


class Migrate(Command):
    async def handle(self):
        from jija_orm.migrator import migrator

        migrator_instance = migrator.Migrator()
        await migrator_instance.migrate()




from jija.config import base, fields


class NetworkConfig(base.Config):
    HOST = fields.CharField(default='0.0.0.0')
    PORT = fields.IntegerField(default=8080)

    def __init__(self, *, host=None, port=None):
        super().__init__(host=host, port=port)
